


def exo2():
    """
    Proposez une fonction fusion(d1, d2) qui renvoie la fusion de deux dictionnaires sans modifier aucun des deux.

    Je propose trois fonctions équivalentes, qui montrent comment réduire le nombre de lignes. Ce n'est pas le seul avantage : En effet, l'implémentation interne de Python fait des optimisations sur les méthodes des structures usuelles.
      - Une liste par compréhension sera plus rapide qu'une liste
      - De même un itérateur (Objet qu'on a vaguement évoqué dans les TD précédents, mais qu'il n'est pas nécessaire de bien connaître au delà de son existence) sera plus rapide.
      - Et si on prend un dictionnaire d (respectivement un ensemble (set) s) d.update(Quelquechose) sera plus rapide que d = {}; for k, v in Quelquechose.items(): d[k] = v;
    """

    def fusion1(d1, d2):
        """
        On crée un dictionnaire vide qu'on remplit.
        """
        d = {}
        for k, v in d1.items():
            d[k] = v
        for k, v in d2.items():
            d[k] = v
        return d


    def fusion2(d1, d2):
        """
        Les boucles for de la fonction précédente se réecrivent aisément à l'aide de la méthode built- in update.
        """
        d = {}
        d.update(d1)
        d.update(d2)
        return d


    def fusion(d1, d2):
        """
        On remarque que les deux premières lignes de la fonction précédente se factorisent en utilisant la méthode copy de la structure dictionnaire.
        """
        d = d1.copy()
        d.update(d2)
        return d
    
    a = {1 : 1, 2 : 3, 5 : 1}
    b = {0 : 3, 6 : 3, 2 : 1}
    print("Si a = {} et b = {}, le résultat attendu de fusion(a,b) est {}".format(a, b, fusion(a,b)))


def exo3():
    """
    Ecrivez une fonction qui prend comme paramètre une liste x et renvoie un dictionnaire indiquant pour chaque élément de la liste son nombre d'apparitions dans celle-ci.
    """

    def Compte1(x):
        """
        Pour cette première façon de faire, on va écrire une fonction auxiliaire qui compte le nombre d'éléments dans une liste. Puis on va l'utiliser afin de résoudre le problème.
        """

        def Compte_aux(L, y):
            """
            Prend une liste L et un élément y et renvoie le nombre d'occurences de y dans L.
            Pour cela, on initialise un compteur à 0, et on l'incrémente à chaque fois qu'on rencontre y en parcourant L.
            """
            nb = 0
            for i in L:
                if i == y:
                    nb+=1 # C'est la même chose que nb = nb + 1
            return nb

        d = {}
        for y in x:
            d[y] = Compte_aux(x, y)
        return d

    # On peut réécrire la fonction précédente de manière "Pythonesque"
    # (Et un peu plus optimisée) en utilisant la méthode count des listes qui réalise exactement
    # la même chose que Compte_aux

    def Compte2(x):
        return {y: x.count(y) for y in x}


    # Cette façon de faire marche très bien, mais peut-on faire encore mieux ?
    # Tout d'abord, il convient de se demander ce qu'on peut améliorer. 
    # À priori, on ne maîtrise pas la méthode count des listes, mais on se doute qu'elle est très
    # proche de la fonction Compte_aux. On va supposer que c'est le cas.
    # On se rend compte alors que : 
    #        - Pour chaque élément y de la liste x, on parcourt toute la liste x et on compare chaque élément à y.
    #    Combien de comparaisons effectue-t-on ?
    #    Si la liste x est de longueur N, alors pour chaque élément de la liste on réalise N comparaisons.
    #   On a donc un total de N² comparaisons. Et il s'agit des seules opérations "coûteuses". En algorithmique, on dit alors que l'algorithme possède une complexité (temporelle) quadratique (Le temps d'exécution est borné par une fonction quadratique du nombre d'éléments) et on dit que cet algorithme est dans la classe O(N²) ou directement qu'il s'exécute en O(N²).

    #  Si vous avez déjà vu les notations de Landau en maths, c'est exactement la même chose.

    # L'objectif est de faire diminuer ce nombre de comparaisons. Pour ça, on va utiliser la méthode get des dictionnaires, dont on va supposer que son temps d'exécution est constant quel que soit le dictionnaire.

    # Petit rappel (Ou pas) sur la méthode get :

    #       - Si d est un dictionnaire, et k est une potentielle clé
    #       - Alors d.get(k) = 
    #                           - d[k] si k est une vraie clé de d
    #                           - None si k n'est pas une clé de d

    # En particulier, d.get(k) ne renvoie jamais d'erreur.


    # En réalité, cette méthode possède un argument optionnel (Comme pour le slicing des listes qu'on a déjà vu, je vous renvoie au TD1 ou au cours pour ça), qui est une valeur par défaut qui sera renvoyée si la clé n'est pas présente dans le dictionnaire :


    #       - Alors d.get(k, valeur) = 
    #                           - d[k] si k est une vraie clé de d
    #                           - valeur si k n'est pas une clé de d


    # En particulier, d.get(k) = d.get(k, None)


    def Compte2(x):
        """
        On a envie de diminuer significativement le temps d'exécution.
        Pour cela, l'idée est d'essayer de faire un seul parcours de la liste, et de réaliser le décompte "à la volée".
        En réalité, on va mimer la fonction de décompte d'occurences en parallèle pour tous les éléments de la liste. 
        Chaque compteur sera directement stocké dans le dictionnaire.
        On va utiliser la méthode get afin d'initialiser les compteurs,
        et on va incrémenter ce compteur à chaque fois qu'on rencontre l'élément associé :
            - Le compteur associé à l'élément y de x sera exactement d[y].
        """
        d = {}
        for y in x:
            d[y] = d.get(y, 0)+1 # Ici, pas besoin de faire de comparaison, on incrémente directement le compteur
        return d



    # Vérifions qu'on a effectivement amélioré la complexité.
    # Ici, on n'a plus de comparaisons. L'élément intéressant est le nombre de parcours de la liste.
    # Plus exactement, en imaginant un curseur qui se déplacerait dans la liste, combien de positions prend-il au cours de l'exécution de l'algorithme ? (Vérifier que cette caractérisation coïncide avec le nombre de comparaisons dans l'exemple précédent)

    # Le curseur en question est localisé dans la variable y. Cette variable prend une fois, et une seule, chaque valeur de la liste x. Donc au total, y prend N valeurs si x est de longueur N.
    # Puisque l'on a admis que la méthode get récupérait des élément en temps constant K, le temps d'exécution de l'algorithme est inférieur K*N. Autrement dit, l'algorithme est en O(N). On dit qu'il possède une complexité linaire.

    x = [2, 5, 2, 2, 1]
    print("Par exemple, si x = {}, Compte2 renvoie {}".format(x, Compte2(x)))


def exo4():
    """
    Écrivez une fonction qui prend en entrée un dictionnaire et un ensemble et vérifie que toutes les valeurs du dictionnaire sont des éléments de l'ensemble.
    """

    def verifie(dico, ens):
        """
        Cette première façon de faire utilise les méthodes des structures dictionnaire et ensemble. Tout d'abord, on récupère les valeurs du dictionnaire dico, qu'on stocke dans un ensemble.
        Puis on vérifie que cet ensemble est bien un sous ensemble de ens.
        """
        return set(dico.values()).issubset(ens)

    def verifie2(dico, ens):
        """
        On peut aussi utiliser la fonction all :
           - all prend en argument une liste et retourne True ssi toutes les valeurs de la liste s'évaluent à True une fois transformés en booléens, et renvoie False sinon.

        On va donc créer une liste qui contient la réponse à la question "Est-ce que x est dans ens ?" pour toutes les valeurs présentes dans le dictionnaire.
        Attention, on se rappelle qu'une boucle for commence par essayer de transformer en liste (En première approximation) l'objet sur lequel on itère. Or, list(dico) est la liste des clés et pas la liste des valeurs. On va donc utiliser la méthode values des dictionnaires pour récupérer la liste des valeurs.
        """
        return all([x in ens for x in dico.values()])

    a = {1 : 1, 2 : 3, 5 : 1}
    ens1 = {3,1,2}
    ens2 = {1,3}
    ens3 = {1}
    
    print("Si a = {} alors :".format(a))
    
    for ens in [ens1, ens2, ens3]:
        if verifie(a, ens):
            print(" - ça marche pour {}".format(ens))
        else:
            print(" - ça ne marche pas pour {}".format(ens))

def exo5():
    """
    Écrivez une fonction qui vérifie que deux listes contiennent exactement les mêmes éléments.
    """
    def memesElements(l1, l2):
        return set(l1) == set(l2)
    

def exo6(Laurent=False):
    """
    Exo sur les graphes
    """
    
    def isGraph(g):
        """
        Proposer une fonction qui prend un paramètre g et vérifie que g représente bien un graphe.
        """
        V, E = g
        for e in E:
            if len(e) != 2:
                return False
            else:
                if e[0] not in V or e[1] not in V:
                    return False
        return True


    g1 = ({0,1,2,3},{(0,1),(2,1),(0,2),(2,3),(3,2),(3,3)})

    h = ({0,1,2,3},{(0,1),(2,1),(0,2),(2,3),(3,2),(3,3),(1,0)})

    g2 = ({0,1,2,3},{(0,1),(2,1),(0,2),(2,3),(3,2),(3,5)})
    #g2 n'est pas un graphe

    g3 = ({0,1,2,3},{(0,1,3),(2,1),(0,2),(2,3),(3,2),(3,3)})
    
    for gr in [g1, g2, g3]:
        print("La structure suivante : {}".format(gr))
        if isGraph(gr):
            print("est un graphe !")
        else:
            print("n'est pas un graphe !")
        


    """
    Correction de Laurent, corrigée ;)
    """
    def verif(g):
        return all({x in g[0] and len(a) == 2 for a in g[1] for x in a})

    if Laurent:
        print(verif(g1))
        print(verif(g2))
        print(verif(g3))


    def isCycle(s, g):
        """
        Teste si s est un cycle de g
        """
        assert isGraph(g)
        if not s:
            return False # Pas de cycle vide
        if len(s)>1:
            return all({(s[i], s[i+1]) in g[1] for i in range(len(s)-1)}) and s[0] == s[-1]
        else:
            return s[0] in g[0] and (s[0], s[0]) in g[1]
    
    cycles = [
        [],
        [3],
        [2,3,2],
        [0,1,2],
        [3,2,1,3],
        [0,1,2,0],
        ]
    print("\n\nOn considère le graphe g = {}".format(g1))
    for cycle in cycles:
        if isCycle(cycle, g1):
            print("{} est un cycle de g".format(cycle))
        else:
            print("{} n'est pas un cycle de g".format(cycle))

    print("\n\nDico")
    
    def dico(g):
        """
        Prend en argument un graphe g et retourne le dictionnaire d tel que :
            - Les clefs sont les sommets
            - La valeur associée à une clé s est l'ensemble des sommets adjacents
        """
        assert isGraph(g)
        d = {i: set() for i in g[0]}
        for e in g[1]:
            d[e[0]].add(e[1])
        return d


    print(dico(g1))


    print("\n\nDico2")
    
    def dico2(g):
        """
        Prend en argument un graphe g et retourne le dictionnaire d tel que :
            - Les clefs sont les sommets
            - La valeur associée à une clé s est l'ensemble des sommets adjacents
        """
        d = {i: set() for i in g.keys()}
        for i in g.keys():
            for s in g[i]:
                d[i].update(g[s])
        return d

    print(dico2(dico(g1)))


    print("\n\nAccesibles")


    def dfs(g, s):
        """
        Parcourt en profondeur du graphe à partir d'un sommet s et renvoit la liste des sommets accessibles depuis s.
        """
        dejaVu = {x: False for x in g[0]}
        accessibles = set()
        d = dico(g)
        Avisiter = [s]
        while Avisiter: # Tant qu'il reste des sommets à voir
            x = Avisiter.pop()
            if not dejaVu[x]:
                dejaVu[x] = True
                voisins = d[x]
                Avisiter.extend(voisins)
                accessibles.update(voisins)
        return accessibles
                
        
    def accessibles(g):
        """
        Construit un dictionnaire tel que :
            - Les clés sont les sommets.
            - La valeur associée à une clé s est l'ensemble des sommets accessibles depuis s.
        """
        return {i: dfs(g, i) for i in g[0]}

    print(accessibles(g1))
    print(accessibles(h))


    print("\n Cycle Hamiltonien\n")

    def hamiltonien(g):
        """
        Teste si le graphe donné en argument possède un circuit hamiltonien.
        """
        d = dico(g)
        # s = list(g[0]).pop()
        # S'il existe un cycle hamiltonien, on peut le faire commencer par n'importe quel sommet.
        cycle = []
        # sommets = {s}

        """
        Idée : On va parcourir tous les sommets et faire un backtracking.
        Pour chaque sommet, on teste si on peut l'ajouter au cycle. 
           - Si non, on passe au sommet suivant.
           - Si oui, on le rajoute au cycle et on essaye de voir si ça nous mène à une solution.
        """

        def peutEtreAjoute(v, cycle, d):
            """
            Teste si le sommet v peut-être ajouté au cycle : 
            Pour cela il faut qu'il soit adjacent au dernier sommet du cycle en construction,
            et qu'il ne soit pas déjà dans le cycle.
            Prend en arguments :
               - Un sommet à tester
               - Un graphe sous forme de dictionnaire
               - 
            """
            return v in d[cycle[-1]] and v not in cycle

        def aux_hamiltonien(d, cycle):
            """
            Arguments :
               - d = graphe sous forme de dictionnaire.
               - cycle = Une liste représentant un cycle courant.
            Fonction auxiliaire récursive qui réaise le backtracking.
            Cas de base : Si le cycle contient tous les sommets, on est bon !
            Cas de récurrence : On teste tous les sommets. 
                                Si on peut l'ajouter, on le fait. Deux cas se produisent alors :
                                      - Soit c'est effectivement un sommet qui mène à une solution. 
                                        C'est testé par un appel récursif.
                                      - Soit il ne mène pas à une solution et on backtrack.
                                Si on ne peut pas l'ajouter, on passe au sommet suivant.
                                Si on a testé tous les sommets, alors le graphe n'est pas hamiltonien.
            """
            
            if len(cycle) == len(d.keys()):
                return cycle
            else:
                return []

            for v in d.keys():
                if peutEtreAjoute(v, cycle, d):
                    cycle.append(v)
                    hamCycle = aux_hamiltonien(d, cycle)
                    if hamCycle:
                        return hamCycle
                    else:
                        cycle.pop(v)
            return []

        hamCycle = aux_hamiltonien(d, cycle)
        if hamCycle:
            if hamCycle[0] in d[cycle[-1]]:
                hamCycle.append(hamCycle[0])
                print("Le graphe {} est bien hamiltonien. Voici un circuit hamiltonien : {}".format(g, hamCycle))
            else:
                print("Le graphe {} possède un chemin hamiltonien mais pas de cycle hamiltonien. Voici un tel chemin : {}".format(g, hamCycle))
        else:
            print("Le graphe {} n'est pas hamiltonien".format(g))
        
    hamiltonien(g1)
    hamiltonien(h)

    if Laurent:
        print("\n\nSolution de Laurent")
        # question 3
        def dicoL(g) :
            return {s:{t for (u,t) in g[1] if u==s} for s in g[0]}

        print(dicoL(g1))


        # question 4
        def dico2L(d) :
            return {s:{t for u in d[s] for t in d[u]} for s in d}

        print(dico2L(dicoL(g1)))


        # question 5
        def accessiblesL(g) :
            d = dico(g)
            d2 = {}
            d3 = d
            while d2!=d3:
                d2 = dico2(d3)
                d3= {s:{t for t in d3[s]|d2[s]} for s in d}
            return(d3)

        print("\nQuestion 5")
        print(accessiblesL(g1))
        print(accessiblesL(h))
